#
# RCSJ model of a DC SQUID, including Johnson noise in junction shunt resistances.
#
# Classes are provided for SQUIDs with different current-phase relations (CPRs) in their junctions:
# "RCSJ_SQUID" - Junctions have the standard, sinusoidal CPR.
# "RCSJ_SGS_SQUID" - Junctions have a sawtooth CPR.
# "RCSJ_short_SQUID" - Junctions have a short SNS CPR, with a shape defined by the transmission of conducting channels.
#
# Equations and variable names follow definitions in
# "The SQUID Handbook Vol. 1: Fundamentals and Technology of SQUIDs and SQUID Systems", 
# J. Clarke and A. I. Braginski, (John Wiley & Sons, 2006)
#

import time
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import numba

from scipy.integrate import solve_ivp
from scipy.signal import welch
from numba import jit

if int(numba.__version__.split('.')[1]) > 50:
    from numba.experimental import jitclass
else:
    from numba import jitclass

# Specification of field types for jitclass
spec_dU_dt_calculator = [
    ('i', numba.float64),
    ('phi_a', numba.float64),
    ('beta_C', numba.float64),
    ('beta_L', numba.float64),
    ('alpha_I', numba.float64),
    ('alpha_R', numba.float64),
    ('alpha_C', numba.float64),
    ('alpha_L', numba.float64),
    ('noise_gamma', numba.float64),
    ('_noise_amplitude', numba.float64),
    ('iN_taus', numba.float64[:]),
    ('iN1', numba.float64[:]),
    ('iN2', numba.float64[:]),
    ('iN_tau_step', numba.float64)
]

@jitclass(spec_dU_dt_calculator)
class dU_dt_calculator:
    # Class to calculate first and second derivatives, with respect to time, of the phase differences
    # across both junctions in the DC SQUID in the case where the current phase relationship (CPR) of 
    # the junctions is sinusoidal.
    #
    #    'delta1' - phase different across junction 1
    #    'delta2' - phase different across junction 2
    # 'delta1_dt' - rate of change of phase different across junction 1
    # 'delta2_dt' - rate of change of phase different across junction 2

    def __init__(self, beta_C, beta_L, alpha_I=0.0, alpha_R=0.0, alpha_C=0.0, alpha_L=0.0, noise_gamma=0.0):
        self.beta_C = beta_C
        self.beta_L = beta_L
        self.alpha_I = alpha_I
        self.alpha_R = alpha_R
        self.alpha_C = alpha_C
        self.alpha_L = alpha_L
        self.noise_gamma = noise_gamma
       
        # Default time step for randomising noise currents
        self.iN_tau_step = 0.1

    def setup_noise(self, tau_start, tau_end):
        self._noise_amplitude = np.sqrt(2.0 * self.noise_gamma / self.iN_tau_step)
        tau_init = tau_start + (self.iN_tau_step*np.random.normal()/4.0)
        self.iN_taus = np.arange(tau_init, tau_end+self.iN_tau_step, self.iN_tau_step)

        if self.noise_gamma == 0.0:
            self.iN1 = np.array([0.0 for tau in self.iN_taus])
            self.iN2 = np.array([0.0 for tau in self.iN_taus])
        else:
            self.iN1 = np.array([self._noise_amplitude * np.random.normal() for tau in self.iN_taus])
            self.iN2 = np.array([self._noise_amplitude * np.random.normal() for tau in self.iN_taus])
        
    def _CPR_norm(self, delta):
        return np.sin(delta)
    
    def calculate(self, tau, U):
        delta1 = U[0]       # Phase difference across junction 1
        delta1_dt = U[1]    # First derivative of delta1 with respect to time
        delta2 = U[2]       # Phase difference across junction 2
        delta2_dt = U[3]    # First derivative of delta2 with respect to time
        
        id = np.searchsorted(self.iN_taus, tau)
        
        cpl = ((delta2 - delta1 - (2.0*np.pi*self.phi_a)) / (np.pi*self.beta_L))
        
        # Calculate second derivates of delta1 and delta2
        delta1_d2t = ((self.i*(1.0+self.alpha_L)/2.0) + cpl - self.iN1[id] - (self._CPR_norm(delta1)*(1.0-self.alpha_I)) - (delta1_dt*(1.0-self.alpha_R))) / (self.beta_C*(1.0-self.alpha_C))
        delta2_d2t = ((self.i*(1.0-self.alpha_L)/2.0) - cpl - self.iN2[id] - (self._CPR_norm(delta2)*(1.0+self.alpha_I)) - (delta2_dt*(1.0+self.alpha_R))) / (self.beta_C*(1.0+self.alpha_C))

        return delta1_dt, delta1_d2t, delta2_dt, delta2_d2t

@jitclass(spec_dU_dt_calculator)
class dU_dt_calculator_SGS:
    # Class to calculate first and second derivatives, with respect to time, of the phase differences
    # across both junctions in the DC SQUID in the case where the current phase relationship (CPR) of 
    # the junctions is a 2-pi periodic sawtooth.
    
    def __init__(self, beta_C, beta_L, alpha_I=0.0, alpha_R=0.0, alpha_C=0.0, alpha_L=0.0, noise_gamma=0.0):
        self.beta_C = beta_C
        self.beta_L = beta_L
        self.alpha_I = alpha_I
        self.alpha_R = alpha_R
        self.alpha_C = alpha_C
        self.alpha_L = alpha_L
        self.noise_gamma = noise_gamma
        
        # Default time step for randomising noise currents
        self.iN_tau_step = 0.1

    def setup_noise(self, tau_start, tau_end):
        self._noise_amplitude = np.sqrt(2.0 * self.noise_gamma / self.iN_tau_step)
        tau_init = tau_start + (self.iN_tau_step*np.random.normal()/4.0)
        self.iN_taus = np.arange(tau_init, tau_end+self.iN_tau_step, self.iN_tau_step)

        if self.noise_gamma == 0.0:
            self.iN1 = np.array([0.0 for tau in self.iN_taus])
            self.iN2 = np.array([0.0 for tau in self.iN_taus])
        else:
            self.iN1 = np.array([self._noise_amplitude * np.random.normal() for tau in self.iN_taus])
            self.iN2 = np.array([self._noise_amplitude * np.random.normal() for tau in self.iN_taus])
        
    def _CPR_norm(self, delta):
        return (np.mod(delta - np.pi, 2.0*np.pi) / np.pi) - 1.0
    
    def calculate(self, tau, U):
        delta1 = U[0]       # Phase difference across junction 1
        delta1_dt = U[1]    # First derivative of delta1 with respect to time
        delta2 = U[2]       # Phase difference across junction 2
        delta2_dt = U[3]    # First derivative of delta2 with respect to time
        
        id = np.searchsorted(self.iN_taus, tau)
        
        cpl = ((delta2 - delta1 - (2.0*np.pi*self.phi_a)) / (np.pi*self.beta_L))
        
        # Calculate second derivates of delta1 and delta2
        delta1_d2t = ((self.i*(1.0+self.alpha_L)/2.0) + cpl - self.iN1[id] - (self._CPR_norm(delta1)*(1.0-self.alpha_I)) - (delta1_dt*(1.0-self.alpha_R))) / (self.beta_C*(1.0-self.alpha_C))
        delta2_d2t = ((self.i*(1.0-self.alpha_L)/2.0) - cpl - self.iN2[id] - (self._CPR_norm(delta2)*(1.0+self.alpha_I)) - (delta2_dt*(1.0+self.alpha_R))) / (self.beta_C*(1.0+self.alpha_C))

        return delta1_dt, delta1_d2t, delta2_dt, delta2_d2t

spec_dU_dt_calculator_short = [
    ('i', numba.float64),
    ('phi_a', numba.float64),
    ('beta_C', numba.float64),
    ('beta_L', numba.float64),
    ('alpha_I', numba.float64),
    ('alpha_R', numba.float64),
    ('alpha_C', numba.float64),
    ('alpha_L', numba.float64),
    ('noise_gamma', numba.float64),
    ('_noise_amplitude', numba.float64),
    ('iN_taus', numba.float64[:]),
    ('iN1', numba.float64[:]),
    ('iN2', numba.float64[:]),
    ('iN_tau_step', numba.float64),
    ('tn', numba.float64[:]),
    ('norm', numba.float64),
    ('skewness', numba.float64),
    ('almost_zero', numba.float64)
]
    
@jitclass(spec_dU_dt_calculator_short)
class dU_dt_calculator_short:
    # Class to calculate first and second derivatives, with respect to time, of the phase differences
    # across both junctions in the DC SQUID in the case where the current phase relationship (CPR) of 
    # the junctions is given by the form for short SNS junctions with one or more conducting channels 
    # with different transmission probability. The channel transmissivities are passed as the list 'tn'
    
    def __init__(self, tn, beta_C, beta_L, alpha_I=0.0, alpha_R=0.0, alpha_C=0.0, alpha_L=0.0, noise_gamma=0.0):
        self.beta_C = beta_C
        self.beta_L = beta_L
        self.alpha_I = alpha_I
        self.alpha_R = alpha_R
        self.alpha_C = alpha_C
        self.alpha_L = alpha_L
        self.noise_gamma = noise_gamma
        
        self.tn = tn
        self.almost_zero = 1e-6
        self.norm = 1.0
        
        # Calculate normalisation constant and skewness of the CPR
        xs = np.arange(-np.pi, np.pi, 2.0*np.pi/1e4)
        ys = np.array([self._CPR_norm(delta) for delta in xs])
        self.norm = np.max(ys)
        max_x = xs[np.where(ys == np.max(ys))[0][0]]
        self.skewness = (max_x - (np.pi/2.)) * 2. / np.pi
        
        # Default time step for randomising noise currents
        self.iN_tau_step = 0.1

    def setup_noise(self, tau_start, tau_end):
        self._noise_amplitude = np.sqrt(2.0 * self.noise_gamma / self.iN_tau_step)
        tau_init = tau_start + (self.iN_tau_step*np.random.normal()/4.0)
        self.iN_taus = np.arange(tau_init, tau_end+self.iN_tau_step, self.iN_tau_step)

        if self.noise_gamma == 0.0:
            self.iN1 = np.array([0.0 for tau in self.iN_taus])
            self.iN2 = np.array([0.0 for tau in self.iN_taus])
        else:
            self.iN1 = np.array([self._noise_amplitude * np.random.normal() for tau in self.iN_taus])
            self.iN2 = np.array([self._noise_amplitude * np.random.normal() for tau in self.iN_taus])
        
    def _CPR_norm(self, delta):
        i = 0.0
        for t in self.tn:
            if t >= 1.0 - self.almost_zero:
                t = 1.0 - self.almost_zero
                
            if t <= self.almost_zero:
                t = self.almost_zero
                
            i += t * np.sin(delta) / np.sqrt(1. - (t*((np.sin(delta/2.))**2.0)))
            
        return i / self.norm
    
    def calculate(self, tau, U):
        delta1 = U[0]       # Phase difference across junction 1
        delta1_dt = U[1]    # First derivative of delta1 with respect to time
        delta2 = U[2]       # Phase difference across junction 2
        delta2_dt = U[3]    # First derivative of delta2 with respect to time
        
        id = np.searchsorted(self.iN_taus, tau)
        
        cpl = ((delta2 - delta1 - (2.0*np.pi*self.phi_a)) / (np.pi*self.beta_L))
        
        # Calculate second derivates of delta1 and delta2
        delta1_d2t = ((self.i*(1.0+self.alpha_L)/2.0) + cpl - self.iN1[id] - (self._CPR_norm(delta1)*(1.0-self.alpha_I)) - (delta1_dt*(1.0-self.alpha_R))) / (self.beta_C*(1.0-self.alpha_C))
        delta2_d2t = ((self.i*(1.0-self.alpha_L)/2.0) - cpl - self.iN2[id] - (self._CPR_norm(delta2)*(1.0+self.alpha_I)) - (delta2_dt*(1.0+self.alpha_R))) / (self.beta_C*(1.0+self.alpha_C))

        return delta1_dt, delta1_d2t, delta2_dt, delta2_d2t

class RCSJ_SQUID:
    # Class to solve the coupled differential equations of a DC SQUID 
    # assuming sinusoidal CPRs in its junctions (using the class 'dU_dt_calculator' above)
    
    _deltas = [0.0, 0.0, 0.0, 0.0]      # Current values of delta1, delta1_dt, delta2, delta2_dt
                                        # delta1(2) is the phase difference across junction 1(2)
                                        # delta1(2)_dt is the first derivateive of delta1(2) with respect to time
    
    _settling_time = 2e3                # ... integrate for this many time units before starting to calculate time-averaged voltage
    _integration_time = 4e4             # ... then integrate over this many time units to calculate time-averaged voltage,
    _integration_time_step = 1e-2       # ... with this step size
    _noise_fft_length = 2e3             # Length (in time units) of the FFT window used to calculate voltage noise
    
    debug_tau = []
    debug_delta1_dt = []
    debug_delta2_dt = []
    
    def __init__(self, beta_C, beta_L, alpha_I=0.0, alpha_R=0.0, alpha_C=0.0, alpha_L=0.0, noise_gamma=0.0):        
        # Select function to calculate time derivatives of delta1 & delta2
        self.dU_dt = dU_dt_calculator(beta_C, beta_L, alpha_I, alpha_R, alpha_C, alpha_L, noise_gamma)
    
    def avg_voltage(self, i, phi_a, calculate_noise=False, debug=[]):
        # Calculate the time-averaged voltage across the SQUID
        # If the current 'i' is in the list 'debug' then show debug information for this value of current.
        
        if i in debug:    
            debug = True
            debug_txt = ' [avg_voltage(i={:.3e})] '.format(i)
        else:
            debug = False
            
        if calculate_noise:
            if self._noise_fft_length > self._integration_time:
                self._noise_fft_length = self._integration_time
        
        self.dU_dt.i = i
        self.dU_dt.phi_a = phi_a
        
        # Coerce the externally-applied flux to lie within +/- 0.5
        # Without this, the simulation does not correctly follow the two branches of a hysteretic junction 
        # when |phi_a| > 0.5
        phi_a = np.mod(phi_a+0.5,1.0)-0.5
        
        tau0 = 0.0              # Start time for numerical integration     
        U0 = self._deltas       # Use the previous saved solution as the starting point for numerical integration
        
        if debug:
            print()
            print(debug_txt + 'Starting. delta1 = {:.3e}, delta1_dt = {:.3e}, delta2 = {:.3e}, delta2_dt = {:.3e}'.format(U0[0], U0[1], U0[2], U0[3]))
            
            # Set up debug plots
            fig, axs = plt.subplots(ncols=2, nrows=2, figsize=(20,10))
            
            axs[0,0].grid(True)
            axs[0,1].grid(True)
            axs[1,0].grid(True)
            axs[1,1].grid(True)
            axs[0,0].set_xlabel(r'$\tau$')
            axs[0,1].set_xlabel(r'$\tau$')
            
            self.debug_tau = []
            self.debug_delta1_dt = []
            self.debug_delta2_dt = []
        
        # Integrate for '_settling_time + _integration_time'
        tau = np.arange(tau0, tau0+self._settling_time+self._integration_time, self._integration_time_step)
        split_id = int(self._settling_time / self._integration_time_step)

        start = time.time()
        self.dU_dt.setup_noise(tau[0], tau[-1])
        solver = solve_ivp(self.dU_dt.calculate, [tau[0], tau[-1]], U0, t_eval=tau, method='RK45', rtol=1e-5, atol=1e-7)
        Us = solver.y
        end = time.time()
        
        U_final = Us[:,-1] # Save the solution at the last time
        
        # Calculate the average voltage, ignoring the first '_settling_time' time units of the result
        voltages = [0.5*(((1.0+self.dU_dt.alpha_L)*d1_dt) + ((1.0-self.dU_dt.alpha_L)*d2_dt)) for d1_dt, d2_dt in zip(Us[1,split_id:], Us[3,split_id:])]
        avg_voltage = np.mean(voltages)
        
        # Calculate low frequency voltage noise
        if calculate_noise:
            noise_N = int(self._noise_fft_length / self._integration_time_step)
            noise_psd_f = np.zeros(int((noise_N / 2) + 1))
            noise_psd_v = np.zeros(int((noise_N / 2) + 1))
            
            # Calculate the spectral density of the calculated voltages vs. time, divided into 'noise_N' sections.
            # Average the resulting 'noise_N' spectra.
            n = 0
            psd = []
            for id1 in np.arange(0, len(voltages)-noise_N+1, noise_N):
                id2 = id1 + noise_N
                vs = voltages[id1:id2]
                psd = welch(vs, fs=1.0/self._integration_time_step, nperseg=noise_N, scaling='density')
                noise_psd_v += psd[1]
                n += 1

            noise_psd_f = psd[0]
            noise_psd_v /= n
            
            if debug:
                print(debug_txt + 'Averaged {} FFT windows (frequencies {:.3e} to {:.3e}, spacing {:.3e}) for voltage noise spectrum'.format(n, noise_psd_f[0], noise_psd_f[-1], noise_psd_f[1]-noise_psd_f[0]))
        
        if debug:
            # Plot debug data
            print(debug_txt + 'Integrated for {:.3e} time units ({:.2f} s calculation time). Average voltage {:.3e}'.format(self._settling_time+self._integration_time, end-start, avg_voltage))

            axs[0,0].plot(tau[split_id:], Us[0,split_id:] / np.pi, '-', label=r'$\delta_1 / \pi$')
            axs[0,0].plot(tau[split_id:], Us[2,split_id:] / np.pi, '-', label=r'$\delta_2 / \pi$')
            axs[0,1].plot(tau[split_id:], Us[1,split_id:], '-', label=r'$d\delta_1/d\tau$')
            axs[0,1].plot(tau[split_id:], Us[3,split_id:], '-', label=r'$d\delta_2/d\tau$')
            
            self.debug_tau = tau
            self.debug_delta1_dt = Us[1,split_id:]
            self.debug_delta2_dt = Us[3,split_id:]
            
            axs[0,0].legend()
            axs[0,1].legend()
            axs[0,0].set_title(r'$i = ' + '{:.3e}'.format(i) + r'$, $\phi_a = ' + '{}'.format(phi_a) + r'$, $V_\mathrm{avg} = ' 
                             + '{:.3e}'.format(avg_voltage) + '$')
            axs[0,1].set_title(r'$i = ' + '{:.3e}'.format(i) + r'$, $\phi_a = ' + '{}'.format(phi_a) + r'$, $V_\mathrm{avg} = ' 
                             + '{:.3e}'.format(avg_voltage) + '$')
            
            if calculate_noise:
                axs[1,0].loglog(noise_psd_f, noise_psd_v)
            
        
        # Save the final junction state as the starting point for the next calculation
        self._deltas = U_final

        if debug:
            plt.show()
            
            if calculate_noise:
                return avg_voltage, noise_psd_f, noise_psd_v, tau, Us
            else:
                return avg_voltage, tau, Us
        else:
            if calculate_noise:
                return avg_voltage, noise_psd_f, noise_psd_v
            else:
                return avg_voltage

class RCSJ_SGS_SQUID(RCSJ_SQUID):
    # Class to solve the coupled differential equations of a DC SQUID 
    # assuming sawtooth CPRs in its junctions (using the class 'dU_dt_calculator_SGS')
    def __init__(self, beta_C, beta_L, alpha_I=0.0, alpha_R=0.0, alpha_C=0.0, alpha_L=0.0, noise_gamma=0.0):
        self.dU_dt = dU_dt_calculator_SGS(beta_C, beta_L, alpha_I, alpha_R, alpha_C, alpha_L, noise_gamma)

class RCSJ_short_SQUID(RCSJ_SQUID):
    # Class to solve the coupled differential equations of a DC SQUID 
    # assuming short, SNS CPRs in its junctions (using the class 'dU_dt_calculator_short')
    def __init__(self, tn, beta_C, beta_L, alpha_I=0.0, alpha_R=0.0, alpha_C=0.0, alpha_L=0.0, noise_gamma=0.0):
        self.dU_dt = dU_dt_calculator_short(tn, beta_C, beta_L, alpha_I, alpha_R, alpha_C, alpha_L, noise_gamma)
