import time
import sys
import gc
import numpy as np
import matplotlib.pyplot as plt
import RCSJ_SQUID_lib as squid_lib

#
# Calculation of transfer function, noise and sensitivity vs. DC current
#

beta_C = 0.1
beta_L = 1.0
alpha_I = 0.0
alpha_R = 0.0
alpha_C = 0.0
alpha_L = 0.0
epsilon = 0.0
settling_time = 2e3
integration_time = 1e4
noise_bw = 0.01
noise_fft_length = 2e3

noise_gamma = 1e-4
i_axis = np.arange(2.5, -0.04, -0.02)
p_axis = np.array([0.24, 0.25, 0.26])

v_threshold = 1e-6

i_array = np.zeros((p_axis.size, i_axis.size))
p_array = np.zeros((p_axis.size, i_axis.size))
v_array = np.zeros((p_axis.size, i_axis.size))
sv_array = np.zeros((p_axis.size, i_axis.size))
sv_err_array = np.zeros((p_axis.size, i_axis.size))

def calc_line(phi_a):
    # For the given applied flux (phi_a) calculate the time-averaged voltage across a SQUID
    # and the voltage noise power for each DC current in the i_axis list. Return voltages as a list.

    # Create SQUID class with the required current-phase relation
    squid = squid_lib.RCSJ_SQUID(beta_C, beta_L, alpha_I=alpha_I, alpha_R=alpha_R, alpha_C=alpha_C, alpha_L=alpha_L, noise_gamma=noise_gamma)
    #squid = squid_lib.RCSJ_SGS_SQUID(beta_C, beta_L, alpha_I=alpha_I, alpha_R=alpha_R, alpha_C=alpha_C, alpha_L=alpha_L, noise_gamma=noise_gamma)
    #squid = squid_lib.RCSJ_short_SQUID(junction_tn, beta_C, beta_L, alpha_I=alpha_I, alpha_R=alpha_R, alpha_C=alpha_C, alpha_L=alpha_L, noise_gamma=noise_gamma)
    
    
    # Set settling time and integration time for numerical integration and
    # the FFT window length used to determine noise power.
    squid._settling_time = settling_time
    squid._integration_time = integration_time
    squid._noise_fft_length = noise_fft_length
    
    vs = np.zeros(i_axis.size)
    svs = np.zeros(i_axis.size)
    sv_errs = np.zeros(i_axis.size)
    
    for idx, i in enumerate(i_axis):
        start = time.time()
        v, noise_f, noise_v = squid.avg_voltage(i, phi_a, calculate_noise=True)
        gc.collect()
        end = time.time()
        duration = end - start

        noise_bw_id = np.argmin(np.fabs(noise_f - noise_bw))
        sv = np.mean(noise_v[2:noise_bw_id])
        sv_err = np.std(noise_v[2:noise_bw_id]) / np.sqrt(noise_bw_id-2)
        
        print('  i = {:.3e}, phi_a = {:.3e} => v = {:.3e}, sv = ({:.3e} +/- {:.3e}) [{:.2f} s calculation time]'.format(i, phi_a, v, sv, sv_err, duration))
        sys.stdout.flush()

        vs[idx] = v
        svs[idx] = sv
        sv_errs[idx] = sv_err
        
    return vs, svs, sv_errs
    
for m, phi_a in enumerate(p_axis):
    print(' ['  + time.ctime() +  '] phi_a = {:.3e} .......'.format(phi_a))
    sys.stdout.flush()
    p_array[m,:] = phi_a
    i_array[m,:] = i_axis
    v_array[m,:], sv_array[m,:], sv_err_array[m,:] = calc_line(phi_a)
    print(' Done.')
    print(' ')
    sys.stdout.flush()


# ---------- Save results

n = np.shape(v_array)[0]
m = np.shape(v_array)[1]

data = np.concatenate(((p_array[:,0]).reshape((n,1)), v_array), axis=1)
i_axis_top = (np.concatenate((np.array([0]), i_array[0,:]), axis=0)).reshape(1,m+1)
data = np.concatenate((i_axis_top, data), axis=0)
np.savetxt('Sensitivity-v.dat', data, delimiter=',')

data = np.concatenate(((p_array[:,0]).reshape((n,1)), sv_array), axis=1)
i_axis_top = (np.concatenate((np.array([0]), i_array[0,:]), axis=0)).reshape(1,m+1)
data = np.concatenate((i_axis_top, data), axis=0)
np.savetxt('Sensitivity-sv.dat', data, delimiter=',')

data = np.concatenate(((p_array[:,0]).reshape((n,1)), sv_err_array), axis=1)
i_axis_top = (np.concatenate((np.array([0]), i_array[0,:]), axis=0)).reshape(1,m+1)
data = np.concatenate((i_axis_top, data), axis=0)
np.savetxt('Sensitivity-sv_err.dat', data, delimiter=',')


# ---------- Plot results

# Calculate transfer function (TF) by numerical differentiation of voltage data
i_axis = i_array[0,:]
TF_array = np.mean(np.diff(v_array, axis=0) / np.diff(p_array, axis=0), axis=0)
mean_vs = np.mean(v_array, axis=0)
mean_svs = np.mean(sv_array, axis=0)

# Calculate sensitivity vs. i
TF_array_tmp = TF_array.copy()
TF_array_tmp[np.where(TF_array == 0)] = np.nan
sv_array_tmp = mean_svs.copy()
sv_array_tmp[np.where(mean_svs == 0)] = np.nan
Sphi_array = np.array([sv / (tf*tf) for sv, tf in zip(sv_array_tmp, TF_array_tmp)])

# Select sensitivity values that are non-nan and correspond to a DC voltage about the threshold (i.e. where the SQUID is normal)
good_ids = np.nonzero(np.logical_and(np.logical_not(np.isnan(Sphi_array)), mean_vs > v_threshold))
    
fmt = 'C0.-'

figsize=(4.6,2.2)
plt.rc('text', usetex=True)
plt.rc('font', family='serif')

fig1, ax1 = plt.subplots(figsize=figsize)
fig2, ax2 = plt.subplots(figsize=figsize)
fig3, ax3 = plt.subplots(figsize=figsize)

ax1.set_ylabel('$v_\phi$', fontsize=11)
ax2.set_ylabel('$s_v$', fontsize=11)
ax3.set_ylabel('$s_\phi = s_v/v_\phi^2$', fontsize=11)

ax1.set_xlabel('$i$', fontsize=11)
ax2.set_xlabel('$i$', fontsize=11)
ax3.set_xlabel('$i$', fontsize=11)

ax1.grid(True)
ax2.grid(True)
ax3.grid(True)

# Add data to transfer function plot
ax1.plot(i_axis, TF_array, fmt)

# Add data to noise plot
ax2.semilogy(i_axis, mean_svs, fmt)

# Add data to sensitivity plot
ax3.semilogy(i_axis[good_ids], Sphi_array[good_ids], fmt)

xmin = 1.3
ax1.set_xlim([xmin, np.max(i_axis)])
ax2.set_xlim([xmin, np.max(i_axis)])
ax3.set_xlim([xmin, np.max(i_axis)])
ax3.set_ylim([2e-4, 0.03])

ax1.tick_params(axis='both', labelsize=11)
ax2.tick_params(axis='both', labelsize=11)
ax3.tick_params(axis='both', labelsize=11)

plt.tightlayout()

fig1.savefig('Sensitivity_vs_i-vphi.pdf')
fig2.savefig('Sensitivity_vs_i-sv.pdf')
fig3.savefig('Sensitivity_vs_i-sphi.pdf')

plt.show()
