import time
import sys
import gc
import numpy as np
import matplotlib.pyplot as plt
import RCSJ_SQUID_lib as squid_lib

#
# Calculation of time-averaged voltage vs. applied flux & DC current
#

beta_C = 1.0
beta_L = 1.0
alpha_I = 0.0
alpha_R = 0.0
alpha_C = 0.0
alpha_L = 0.0
noise_gamma = 0.0
junction_tn = np.array([0.9706])
settling_time = 2e3
integration_time = 2e3

i_axis = np.arange(3.0, -3.05, -0.05)
p_axis = np.arange(-1.01, 1.03, 0.02)

i_array = np.zeros((p_axis.size, i_axis.size))
p_array = np.zeros((p_axis.size, i_axis.size))
v_array = np.zeros((p_axis.size, i_axis.size))

def calc_line(phi_a):
    # For the given applied flux (phi_a) calculate the time-averaged voltage across a SQUID
    # for each DC current in the i_axis list. Return voltages as a list.

    # Create SQUID class with the required current-phase relation
    #squid = squid_lib.RCSJ_SQUID(beta_C, beta_L, alpha_I=alpha_I, alpha_R=alpha_R, alpha_C=alpha_C, alpha_L=alpha_L, noise_gamma=noise_gamma)
    squid = squid_lib.RCSJ_SGS_SQUID(beta_C, beta_L, alpha_I=alpha_I, alpha_R=alpha_R, alpha_C=alpha_C, alpha_L=alpha_L, noise_gamma=noise_gamma)
    #squid = squid_lib.RCSJ_short_SQUID(junction_tn, beta_C, tn, beta_L, alpha_I=alpha_I, alpha_R=alpha_R, alpha_C=alpha_C, alpha_L=alpha_L, noise_gamma=noise_gamma)
    
    # Set settling time and integration time for numerical integration
    squid._settling_time = settling_time
    squid._integration_time = integration_time
    
    # Calculate time-averaged voltage across the SQUID for DC currents in the i_axis list
    vs = []   
    for i in i_axis:
        start = time.time()
        v = squid.avg_voltage(i, phi_a)
        gc.collect()
        end = time.time()
        duration = end - start
        print('  i = {:.3e}, phi_a = {:.3e} => v = {:.3e} [{:.2f} s calculation time]'.format(i, phi_a, v, duration))
        sys.stdout.flush()
        vs.append(v)
        
    return vs

# Calculate voltages for each applied flux in the p_axis list
for m, phi_a in enumerate(p_axis):
    print(' ['  + time.ctime() +  '] phi_a = {:.3e} .......'.format(phi_a))
    sys.stdout.flush()
    p_array[m,:] = phi_a
    i_array[m,:] = i_axis
    v_array[m,:] = calc_line(phi_a)
    print(' Done.')
    print(' ')
    sys.stdout.flush()

# ---------- Save results

n = np.shape(v_array)[0]
m = np.shape(v_array)[1]
    
data = v_array
data = np.concatenate(((p_array[:,0]).reshape((n,1)), v_array), axis=1)
i_axis_top = (np.concatenate((np.array([0]), i_array[0,:]), axis=0)).reshape(1,m+1)
data = np.concatenate((i_axis_top, data), axis=0)

np.savetxt('v_vs_phi_i.dat', data, delimiter=',')


# ---------- Plot results

fig, ax = plt.subplots(figsize=(3.5, 3.5))

vmax = np.fabs(v_array).max()
vmin = -1*vmax

ax.tick_params(labelsize=11)

cax = ax.pcolormesh(p_array, i_array, v_array, shading='auto', vmin=vmin, vmax=vmax, cmap='bwr', linewidth=0, rasterized=True)
ax.axis([-1, 1, -3, 3])

cbar = fig.colorbar(cax)
cbar.set_label(r'$\left<v\right>$', size=11)
cbar.ax.tick_params(labelsize=11)

ax.set_xlabel(r'$\phi_a$', size=11)
ax.set_ylabel(r'$i$', size=11)
plt.grid(True)

plt.tight_layout()
plt.savefig('v_vs_phi_i.pdf')

plt.show()


