import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

task_ids = [2, 2]
v_threshold = 1e-6
datadirs = ['symmetric_Sensitivity_vs_gamma_beta_C_10e-2_beta_L_10e-1_epsilon_0e+0_int_1e+4', 
            'SGS_symmetric_Sensitivity_vs_gamma_beta_C_10e-2_beta_L_10e-1_epsilon_0e+0_int_1e+4']

fmts = [ 'C0.-', 'C1.--']
labels = ['$\sin(\delta)$', 'sawtooth']

figsize=(4.6,2.0)
plt.rc('text', usetex=True)
plt.rc('font', family='serif')

fig1, ax1 = plt.subplots(figsize=figsize)
fig2, ax2 = plt.subplots(figsize=figsize)
fig3, ax3 = plt.subplots(figsize=figsize)

ax1.set_ylabel('$v_\phi$', fontsize=11)
ax2.set_ylabel('$s_v$', fontsize=11)
ax3.set_ylabel('$s_\phi = s_v/v_\phi^2$', fontsize=11)

ax1.set_xlabel('$i$', fontsize=11)
ax2.set_xlabel('$i$', fontsize=11)
ax3.set_xlabel('$i$', fontsize=11)

ax1.grid(True)
ax2.grid(True)
ax3.grid(True)

for datadir, task_id, fmt, label in zip(datadirs, task_ids, fmts, labels):
    # Load DC voltage data from a previous calculation
    data = np.loadtxt(datadir + '/Sensitivity_vs_gamma-{:05d}-v.dat'.format(task_id), delimiter=',')

    v_array = data[1:,1:]
    p_array = np.zeros(np.shape(v_array))
    i_array = np.zeros(np.shape(v_array))

    for i in range(np.shape(v_array)[1]):
        p_array[:,i] = data[1:,0]

    for i in range(np.shape(v_array)[0]):
        i_array[i,:] = data[0,1:]
        
    # Load voltage noise data from a previous calculation
    data = np.loadtxt(datadir + '/Sensitivity_vs_gamma-{:05d}-sv.dat'.format(task_id), delimiter=',')
    sv_array = data[1:,1:]

    # Calculate transfer function (TF) by numerical differentiation of voltage data
    i_axis = i_array[0,:]
    TF_array = np.mean(np.diff(v_array, axis=0) / np.diff(p_array, axis=0), axis=0)
    mean_vs = np.mean(v_array, axis=0)
    mean_svs = np.mean(sv_array, axis=0)

    # Calculate sensitivity vs. i
    TF_array_tmp = TF_array.copy()
    TF_array_tmp[np.where(TF_array == 0)] = np.nan
    sv_array_tmp = mean_svs.copy()
    sv_array_tmp[np.where(mean_svs == 0)] = np.nan
    Sphi_array = np.array([sv / (tf*tf) for sv, tf in zip(sv_array_tmp, TF_array_tmp)])

    # Select sensitivity values that are non-nan and correspond to a DC voltage about the threshold (i.e. where the SQUID is normal)
    good_ids = np.nonzero(np.logical_and(np.logical_not(np.isnan(Sphi_array)), mean_vs > v_threshold))

    # Add data to transfer function plot
    ax1.plot(i_axis, TF_array, fmt, label=label)
    
    # Add data to noise plot
    ax2.semilogy(i_axis, mean_svs, fmt, label=label)
    
    # Add data to sensitivity plot
    ax3.semilogy(i_axis[good_ids], Sphi_array[good_ids], fmt, label=label)
    
xmin = 1.3
ax1.set_xlim([xmin, np.max(i_axis)])
ax2.set_xlim([xmin, np.max(i_axis)])
ax3.set_xlim([xmin, np.max(i_axis)])
ax3.set_ylim([2e-4, 0.03])

ax1.tick_params(axis='both', labelsize=11)
ax2.tick_params(axis='both', labelsize=11)
ax3.tick_params(axis='both', labelsize=11)

ax1.legend(fontsize=11)

fig1.savefig('sensitivity_comparison-v_phi.pdf', bbox_inches = 'tight')
fig2.savefig('sensitivity_comparison-s_v.pdf', bbox_inches = 'tight')
fig3.savefig('sensitivity_comparison-s_phi.pdf', bbox_inches = 'tight')

